// Copyright (C) 2008 Indrek Mandre <indrek@mare.ee>

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <time.h>

#define JBOUND 0.5        // exchange bound energy
#define KB 1.0            // the Boltzmann's constant
#define TAU 1.0           // time scale factor

#define RNDSEED 1         // random seed

class Grid
{
  struct Cell
  {
    char v;   // spin value, -1 or +1
    char nc;  // neighbouring spins sum
  };

  int LN;
  double muH;
  Cell *_GRID;

  double TRMAP[14]; // transition probability map

  int ebs;        // energy bounds sum
  int sbalance;   // up/down spins sum

  double temp;  // temperature

  inline Cell& get (int plane, int row, int column)
  {
    if ( row == -1 ) row = LN - 1;
    if ( row == LN ) row = 0;
    if ( column == -1 ) column = LN - 1;
    if ( column == LN ) column = 0;
    if ( plane == -1 ) plane = LN - 1;
    if ( plane == LN ) plane = 0;
    return fget(plane, row, column);
  }

  inline Cell& fget (int plane, int row, int column)
  {
    return _GRID[plane * LN * LN  + row * LN + column];
  }

public:
  Grid (int LN = 16, double temp = 0, double muH = 0) :
      LN(LN), _GRID(0) { setup(LN, temp, muH); }
  ~Grid () { delete [] _GRID; }

  // set up the simulation at temperature t
  void setup (int LN = 16, double t = 0, double muH = 0)
  {
    this->LN = LN;
    this->muH = muH;
    delete [] _GRID;
    _GRID = new Cell [LN*LN*LN];
    temp = t;
    sbalance = 0;
    ebs = 0;

    // set all spins to alternating -1/1
    for ( int i = 0; i < LN; i++ )
        for ( int j = 0; j < LN; j++ )
            for ( int k = 0; k < LN; k++ )
                fget(i, j, k).v = muH < 0 ? -1 : 1;

    // calculate contributions from neighbours
    for ( int i = 0; i < LN; i++ )
        for ( int j = 0; j < LN; j++ )
            for ( int k = 0; k < LN; k++ )
                fget(i, j, k).nc =
                    get(i - 1, j, k).v + get(i + 1, j, k).v +
                    get(i, j - 1, k).v + get(i, j + 1, k).v +
                    get(i, j, k - 1).v + get(i, j, k + 1).v;

    // calculate energy bound count
    for ( int i = 0; i < LN; i++ )
        for ( int j = 0; j < LN; j++ )
            for ( int k = 0; k < LN; k++ )
              {
                Cell& c = fget(i, j, k);
                ebs += c.v * c.nc;
                sbalance += c.v;
              }

    reprob ();
  }

  // calculate probabilities for the state transitions
  void reprob ()
  {
    for ( int ncv = -6; ncv <= 6; ncv += 2 )
      {
        int i = ncv + 6;
        // from the positive spin +1
        double dH = -4 * (-JBOUND * ncv) + 2 * muH;
        double prob = exp(-dH/(KB * temp)) / TAU;
        if ( prob > 1.0 / TAU ) prob = 1.0 / TAU;
        TRMAP[i] = prob;

        // from the negative spin -1
        dH = -4 * (-JBOUND * ncv) - 2 * muH;
        prob = exp(-dH/(KB * temp)) / TAU;
        if ( prob > 1.0 / TAU ) prob = 1.0 / TAU;
        TRMAP[i + 1] = prob;
      }
  }

  // change the muH parameter
  void set_muH(double muH)
  {
    this->muH = muH;
    reprob ();
  }

  // toggle the spin of the given cell
  inline void toggle (int plane, int row, int column)
  {
    Cell& c = get(plane, row, column);
    c.v *= -1;
    get(plane - 1, row, column).nc += 2 * c.v;
    get(plane + 1, row, column).nc += 2 * c.v;
    get(plane, row - 1, column).nc += 2 * c.v;
    get(plane, row + 1, column).nc += 2 * c.v;
    get(plane, row, column - 1).nc += 2 * c.v;
    get(plane, row, column + 1).nc += 2 * c.v;
    sbalance += 2 * c.v;
    ebs += 4 * c.v * c.nc;
  }

  // move simulation on by one simulation step
  void step ()
  {
    for ( int i = 0; i < LN; i++ )
        for ( int j = 0; j < LN; j++ )
            for ( int k = 0; k < LN; k++ )
              {
                Cell& c = fget(i, j, k);
                if ( drand48() < TRMAP[c.v * c.nc + 6 +
                    (c.v == -1 ? 1 : 0) ] ) toggle (i, j, k);
              }
  }

  // calculate the energy in the simulation
  double energy ()
  {
    return -JBOUND * ebs - muH * sbalance;
  }

  // calculate the energy contribution from the given cell
  double energy (int plane, int row, int column)
  {
    Cell& c = get(plane, row, column);
    return -JBOUND * c.v * c.nc - muH * c.v;
  }

  inline int spin (int plane, int row, int column)
  {
    return get(plane, row, column).v;
  }

  double magnetization ()
  {
    return (double)sbalance / (LN * LN * LN);
  }

  struct sample_data
  {
    double magnetization;
    double abs_magnetization;
  };

  void sample (sample_data& out, size_t eqc, size_t sampc)
  {
    for ( size_t i = 0; i < eqc; i++ )
        step();
    double m = 0;
    double abs_m = 0;
    for ( size_t i = 0; i < sampc; i++ )
      {
        step();
        m += magnetization();
        abs_m += fabs(magnetization());
      }
    out.magnetization = m / sampc;
    out.abs_magnetization = abs_m / sampc;
  }
};

static void compare_grid(int n)
{
  printf ("Compare grid %d\n", n);
  char fn[128];
  sprintf (fn, "cgrid%d.txt", n);
  FILE *fp = fopen (fn, "w+");
  Grid *grid = new Grid ();
  for ( double t = 3; t < 6.01;
      t += (fabs(t - 4.5) < 0.5) ? 0.025 : 0.05 )
    {
      Grid::sample_data data;
      grid->setup(n, t);
      grid->sample (data, 1000, 1000);
      fprintf (fp, "%f %f\n", t, data.abs_magnetization);
    }
  delete grid;
  fclose (fp);
}

static void compare_hm (double temp)
{
  printf ("M=M(H) / TEMP = %.1f\n", temp);
  char fn[128];
  sprintf (fn, "chm%.1f.txt", temp);
  FILE *fp = fopen (fn, "w+");
  double h = 0;
  Grid *grid = new Grid (20, temp, h);
  Grid::sample_data data;
  grid->sample (data, 1000, 0);
  for ( ; h <= 5.01; h += 0.25 )
    {
      grid->set_muH (h);
      grid->sample (data, 200, 200);
      fprintf (fp, "%f %f\n", h, data.magnetization);
    }
  for ( ; h >= -5.01; h -= 0.25 )
    {
      grid->set_muH (h);
      grid->sample (data, 200, 200);
      fprintf (fp, "%f %f\n", h, data.magnetization);
    }
  for ( ; h <= 5.01; h += 0.25 )
    {
      grid->set_muH (h);
      grid->sample (data, 200, 200);
      fprintf (fp, "%f %f\n", h, data.magnetization);
    }
  delete grid;
  fclose (fp);
}

static void running_time (int n1, int n2)
{
  printf ("Running time...\n");
  FILE *fp = fopen ("rtime.txt", "w+");
  Grid *grid = new Grid ();
  for ( int i = n1; i <= n2; i++ )
    {
      printf ("running_time: %d\n", i);
      Grid::sample_data data;
      grid->setup (i, 0);
      clock_t t1, t2;
      t1 = clock ();
      grid->sample (data, 100000, 0);
      t2 = clock ();
      fprintf (fp, "%d %f\n", i, (double)(t2 - t1)/1000000.0);
    }
  delete grid;
  fclose (fp);
}

static void slice (double temp)
{
  printf ("Running slice...\n");
  FILE *fp = fopen ("slice.inc", "w+");
  Grid *grid = new Grid (30, temp);
  Grid::sample_data data;
  grid->sample (data, 2000, 1000);
  printf ("Magnetization: %f\n", data.magnetization);
  for ( int i = 0; i < 30; i++ )
      for ( int j = 0; j < 30; j++ )
          for ( int k = 0; k < 30; k++ )
            {
              if ( k <= 15 || j <= 15 )
                  fprintf (fp, "sphere{<%d,%d,%d>,0.7 texture{%s}}\n",
                      i, j, k, grid->spin(i, j, k) < 0 ? "SD" : "SU");
            }
  delete grid;
  fclose (fp);
}

int main ()
{
  srand48 (RNDSEED);

  slice (4.3);

  compare_hm (0.0);
  compare_hm (2.0);
  compare_hm (4.5);
  compare_hm (6.0);

  running_time (5, 20);

  compare_grid (10);
  compare_grid (20);
  compare_grid (30);

  return 0;
}


